"""List some information about all filters available in the current directory.

This is a development-time introspection tool that is helpful for getting an overview
of the default stageName and sort values of all filters at a glance. This is especially useful
when looking for the appropriate position to place a new type of filter in.

The reported information is module name, class name, and stageName and sort of each Filter found.

This tool should be run in direct/src/filter with

  python -m ListAllFilters

"""

from __future__ import print_function

import imp, os, re
import importlib
import inspect
import collections

# These must be imported first before importing any filters to avoid dependency errors (FIXME).
#
import FilterPipeline
import FilterManager

import Filter

# The module finder comes from a tool in Python 2.5, by Fredrik Lundh, 2005:
#   http://svn.python.org/projects/python/tags/r252/Doc/tools/listmodules.py

_suffixes = imp.get_suffixes()
def _get_suffix(file):
    for suffix in _suffixes:
        if file[-len(suffix[0]):] == suffix[0]:
            return suffix
    return None

def getmodules(p):
    """Get modules in a given directory p (string, path)."""
    modules = {}
    for f in os.listdir(p):
        f = os.path.join(p, f)
        if os.path.isfile(f):
            m, e = os.path.splitext(f)
            suffix = _get_suffix(f)
            if not suffix:
                continue
            m = os.path.basename(m)
            if re.compile("(?i)[a-z_]\w*$").match(m):
                if suffix[2] == imp.C_EXTENSION:
                    # check that this extension can be imported
                    try:
                        __import__(m)
                    except ImportError:
                        continue
                modules[m] = f
        elif os.path.isdir(f):
            m = os.path.basename(f)
            if os.path.isfile(os.path.join(f, "__init__.py")):
                for mm, f in getmodules(f).items():
                    modules[m + "." + mm] = f
    return modules


# Our strategy is to find all modules, import each of them,
# and find all subclasses of Filter defined in each.

# Modules with these names do not define any filters themselves, but they may import some
# (either directly or via other imports), which would mess up the inspection logic.
#
moduleBlacklist = ["FilterPipeline", "CommonFilters", "FilterStage"]

RawFilterInfo = collections.namedtuple( "RawFilterInfo", ["modName", "clsName", "cls"] )

def getFilters(p):
    """Get Filters defined in a given directory p (string, path)."""
    filterModules = getmodules(p)
    rawFilterInfos = []
    for modName in filterModules.keys():
        if modName in moduleBlacklist:
            continue
        mod = importlib.import_module(modName)
        for clsName in dir(mod):
            cls = getattr(mod, clsName)
            if not inspect.isclass(cls):
                continue
            if issubclass(cls, Filter.Filter):
                rawFilterInfos.append( RawFilterInfo(modName, clsName, cls) )
    return rawFilterInfos


# The filter search will skip trying to instantiate classes with these names.
#
knownAbstractClassNames = ["Filter", "CompoundFilter", "RapidPrototypeCompoundFilter"]

# FIXME: This default list is an instance member of FilterPipeline, filled in __init__(),
# FIXME: and currently there is no way to create a FilterPipeline without having win and cam.
#
knownStages = [ "Preprocess", "SceneOptics", "LensFocus", "LensOpticsEarly", "LensOpticsLate",
                "FilmOrDetector", "Postprocess", "DisplayDevice", "DebugHelpers" ]


PrintableFilterInfo = collections.namedtuple( "PrintableFilterInfo", ["modName", "clsName", "stageName", "sort"] )

def report(sortKeyName="render"):
    """Report Filters in the current directory into stdout.

    The report is printed in a fixed-width format.

    The ordering is controlled by sortKeyName:
      "render"      : order by stageName (in logical order), then the sort value
      "class"       : order by just the class name, globally
      "moduleclass" : order by module name, then by class name within each module

    """

    filterInfos = []
    for rawFilterInfo in getFilters(p="."):
        if rawFilterInfo.clsName in knownAbstractClassNames:
            continue
        instance = rawFilterInfo.cls()
        filterInfos.append( PrintableFilterInfo(rawFilterInfo.modName, rawFilterInfo.clsName,
                                                instance.stageName, instance.sort) )
    headerItem = PrintableFilterInfo("Module", "Class", "stageName", "sort")

    k = [None] + knownStages  # StageInitializer has stageName=None
    sortKeys = { "render"      : lambda item: (k.index(item.stageName), item.sort),
                 "class"       : lambda item: item.clsName,
                 "moduleclass" : lambda item: (item.modName, item.clsName) }

    filterInfos.sort( key=sortKeys[sortKeyName] )
    filterInfos = [headerItem] + filterInfos

    # Perform some fixed-width formatting in the output.
    #
    def columnLength(col, item):
        return len( str(item[col]) )

    maxColumnLengths = []
    for col in xrange(len(headerItem)):  # each item is assumed to have the same number of columns as the header.
        maxColumnLengths.append(  max( map( lambda item : columnLength(col, item), filterInfos ) )  )

    prevStageName = None
    for lineno,item in enumerate(filterInfos):
        isHeaderLine = (lineno == 0)

        columns = []
        for col,text in enumerate(item):
            pad = " " * (maxColumnLengths[col] - columnLength(col, item))
            columns.append( "%s%s" % (text, pad) )
        line = " ".join(columns)  # observe that line has a constant width.

        # When reporting in render order, mark logical stage boundaries
        # (special-casing the header and the initial "None").
        #
        if sortKeyName == "render"  and  not isHeaderLine  and  item.stageName is not None:
            if item.stageName != prevStageName:
                print( "-" * len(line) )
                prevStageName = item.stageName

        print( line )

        # Print separator after header.
        #
        if isHeaderLine:
            print( "=" * len(line) )


def main():
    """Print reports with all three orderings."""

    print("\nAll filters by class name:\n")
    report(sortKeyName="class")
    print("")

    print("\nAll filters by module, then class:\n")
    report(sortKeyName="moduleclass")
    print("")

    print("\nAll filters in render order (with default settings):\n")
    report(sortKeyName="render")
    print("")


if __name__ == '__main__':
    main()

