"""A collection of short, simple filters for FilterPipeline.

More complex filters each have their own module.

"""

from pandac.PandaModules import AuxBitplaneAttrib, Vec4

from Filter import Filter, TextureInfo

###################################################
# Special magic filter to initialize a render pass
###################################################

class StageInitializer(Filter):
    """This filter simply initializes pixcolor from the current pixel in the pipeline stage input color texture.

    It is automatically inserted into each FilterStage created during FilterPipeline.reconfigure().

    Unlike other filters, the properties of StageInitializer are not meaningful,
    and the filter is owned by (the only is reference stored in) the FilterStage it resides in.
    Hence, during each reconfigure of the pipeline, old StageInitializers are destroyed,
    and new ones created (because FilterStages are).

    The sort value must be -1, which places StageInitializer before any other filters in the stage.
    Only StageInitializer is allowed to have a sort value of -1.

    """
    def __init__(self, **kwargs):
        super(StageInitializer, self).__init__(**kwargs)

        # Unlike usual, this filter doesn't want the compositing fshader
        # to supply the current value of pixcolor to the filter function
        # (because the purpose of StageInitializer is to
        #  determine and assign the initial value of pixcolor).
        #
        # For more information, see Filter.__init__().
        #
        self._needPixcolor = False

    def reset(self):
        super(StageInitializer, self).reset()  # reset inherited properties
        self.sort        = -1     # Initialization goes first before any other filter
                                  # in the same FilterStage.

        self.stageName   = None   # This filter is special; the value of stageName is not used.
                                  # Instead, each created render pass gets one copy of the filter.

        self.isMergeable = None   # The initialization filter is only added after the logical stage
                                  # into render pass merging process is complete. The value of isMergeable
                                  # is ignored.

    def setup(self, phase):
        if phase == "pipeline":
            # The compositing shader needs the color texture to initialize the output color.
            self.registerInputTexture(texName="color")

    def synthesize(self):
        # Registered textures are accessed with self.getTextureInfo(texName).
        # This ensures that the correct definition is returned.
        #
        # See Filter.TextureInfo for the available fields.
        #
        # pixcolor is the input/output variable for the color of the current pixel.
        #
        txcolor = self.getTextureInfo("color")
        return ("stageInitializer",
                "float4 pixcolor = tex2D(%(k_txcolor)s, %(texcoord_txcolor)s);\n" % { "k_txcolor" : txcolor.varname,
                                                                                      "texcoord_txcolor" : txcolor.texcoord },
                "// initialize pixcolor\n")


###################################################
# Actual regular MiscFilters
###################################################

class Tint(Filter):
    """Color tint / fade-to-color filter.

    In tint mode [default]:

      The color of each pixel is tinted (multiplied) by the specified constant color.
      The result is saturated to [0,1] in each component.

      The tinted pixel then is blended with the original one with the strength
      specified by blend.

    In fade mode:

      Each pixel is blended toward the specified constant color.
      The parameter blend controls the strength.

    This filter can be used e.g. to fade the screen to black (color=(0,0,0,1))
    by dynamically updating the blend parameter in your own update task
    so that it gradually goes from 0 to 1. (Make your task sort 48 or earlier
    to ensure it runs before filter reconfigure/update and frame redraw (igLoop).)

    The fade mode also allows fading toward colors other than black
    (e.g. color=(1,1,1,1) to fade to white).

    In tint mode, fading the color to (255,255,255,1) gives another kind of
    fade to white (using the fact that there are 8 bits per color channel).

    Parameters:

      mode  = string, one of:
                "tint" = tint (multiply) by the specified color.
                "fade" = fade to the specified color.

      color = (R,G,B,A) tuple; tint color. To make a component no-op in tint mode,
              set it to 1. Usually setting the alpha component to 1 is a good idea
              regardless of the mode.

      blend = strength of blending between the original and processed color.
              1.0 means fully tinted or faded, 0.0 does nothing.
              Behaves linearly.

    """
    def __init__(self, **kwargs):
        super(Tint, self).__init__(**kwargs)

    def reset(self):
        super(Tint, self).reset()  # reset inherited properties
        self.sort  = 90  # late within the pipeline stage
        self.stageName = "Postprocess"
        self.color = (0.5,0.5,0.5,1)  # default is darken by 50%
        self.blend = 1.0
        self.mode  = "tint"  # compile-time parameter

    def setup(self, phase):
        if phase == "pipeline":
            # We do not need to register the "color" texture, as our code does not access it explicitly,
            # and we do not need texture coordinates, either.
            #
            # The StageInitializer sets up the initial value of pixcolor automatically.
            # It is thus better to refer to pixcolor for the original color of the pixel being processed,
            # to save one texture lookup (and unnecessary parameters to our Cg function generated
            # in synthesize()).
            #
            self.registerCustomInput(inputType="float4", inputName="k_tint_color")
            self.registerCustomInput(inputType="float",  inputName="k_tint_blend")

    def synthesize(self):
        code = ""
        if self.mode == "tint":
            code += "// Compile-time mode: color tint\n"
            code += "float4 tinted_color = saturate(pixcolor * %(k_tint_color)s);\n" % { "k_tint_color" : self.getMangledName("k_tint_color") }
            code += "pixcolor = lerp(pixcolor, tinted_color, %(k_tint_blend)s);\n" % { "k_tint_blend" : self.getMangledName("k_tint_blend") }
        elif self.mode == "fade":
            code += "// Compile-time mode: fade-to-color\n"
            code += "pixcolor = lerp(pixcolor, %(k_tint_color)s, %(k_tint_blend)s);\n" % \
                        { "k_tint_color" : self.getMangledName("k_tint_color"),
                          "k_tint_blend" : self.getMangledName("k_tint_blend") }
        else:
            raise ValueError("Tint: invalid mode '%s'; valid: 'tint', 'fade'" % self.mode)
        return ("tint", code, "// Color tint / fade-to-color filter\n")


    @property
    def color(self):
        """Tint color as (R,G,B,A) float tuple, each component in [0,1]."""
        return self._color
    @color.setter
    def color(self, value):
        self._color = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("tint_color"), value )

    @property
    def blend(self):
        """Effect strength, float. 1 is full strength, 0 is off."""
        return self._blend
    @blend.setter
    def blend(self, value):
        self._blend = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("tint_blend"), value )

    @property
    def mode(self):
        """Effect mode, either 'tint' or 'fade'."""
        return self._mode
    @mode.setter
    def mode(self, value):
        if (not hasattr(self, "_mode")  or  value != self._mode)  and  self.pipeline is not None:
            self.pipeline._needsCompile = True
        self._mode = value


class Scanlines(Filter):
    """Filter that simulates CRT TV scanlines by darkening alternating rows of the picture."""

    FIELD_TOP    = 0
    FIELD_BOTTOM = 1

    def __init__(self, **kwargs):
        super(Scanlines, self).__init__(**kwargs)

    def reset(self):
        super(Scanlines, self).reset()
        self.sort  = 95
        self.stageName = "DisplayDevice"
        self.field = Scanlines.FIELD_TOP
        self.blend = 0.5
        self.line_thickness = 1
        self.dynamic = False

    def setup(self, phase):
        if phase == "pipeline":
            self.registerInputTexture(texName="color")
            self.registerCustomInput(inputType="int",   inputName="k_scanlines_field")
            self.registerCustomInput(inputType="float", inputName="k_scanlines_blend")
            self.registerCustomInput(inputType="int",   inputName="k_scanlines_thickness")

            # This filter, when in dynamic mode, requires updating the "field" input at each frame.
            #
            # We always add the task so that "dynamic" will behave as expected if switched on/off
            # on the fly.
            #
            self.registerUpdatable()

    def synthesize(self):
        txcolor = self.getTextureInfo("color")
        code  = "if (int(%(texcoord_txcolor)s.y / %(texpix_txcolor)s.y)/%(k_scanlines_thickness)s %% 2 != %(k_scanlines_field)s)\n" % \
                    { "texcoord_txcolor" : txcolor.texcoord,
                      "texpix_txcolor" : txcolor.texpix,
                      "k_scanlines_thickness" : self.getMangledName("k_scanlines_thickness"),
                      "k_scanlines_field" : self.getMangledName("k_scanlines_field") }
        code += "    pixcolor.rgb = (1.0 - %(k_scanlines_blend)s) * pixcolor.rgb;\n" % { "k_scanlines_blend" : self.getMangledName("k_scanlines_blend") }
        return ("scanlines", code, "// CRT TV scanlines filter\n")

    def update(self):
        # Note that we use the value of an UTParam (meant for update() to read)
        # to decide whether to update an RTParam (shader input).
        #
        if self.dynamic:
            self.field = 1 - self.field


    @property
    def field(self):
        """Which field to keep, either Scanlines.FIELD_TOP or Scanlines.FIELD_BOTTOM."""
        return self._field
    @field.setter
    def field(self, value):
        self._field = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("scanlines_field"), value )

    @property
    def blend(self):
        """Effect strength (float). 1 is fully scanlined, 0 does nothing."""
        return self._blend
    @blend.setter
    def blend(self, value):
        self._blend = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("scanlines_blend"), value )

    @property
    def line_thickness(self):
        """Thickness of each scanline in pixels (integer)."""
        return self._line_thickness
    @line_thickness.setter
    def line_thickness(self, value):
        self._line_thickness = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("scanlines_thickness"), value )

    @property
    def dynamic(self):
        """Dynamic mode (bool). When enabled, the 'field' parameter is toggled at each frame rendered. With a steady refresh rate synced to display refresh, this creates an old CRT TV appearance, but the effect is very sensitive to even small unevenness in the framerate. (Hence, default False.)"""
        return self._dynamic
    @dynamic.setter
    def dynamic(self, value):
        # This parameter is queried in update(); no need to do anything here except store the new value.
        self._dynamic = value


class ViewGlow(Filter):
    """View the glow map by overwriting the red color channel with the glow data. This may help in debugging your scene."""

    def __init__(self, **kwargs):
        super(ViewGlow, self).__init__(**kwargs)

    def reset(self):
        super(ViewGlow, self).reset()
        self.stageName = "DebugHelpers"  # this is a debug helper filter which should go last
        self.sort = 99  # last thing in the stage

    def setup(self, phase):
        if phase == "pipeline":
            self.requireAuxBits(bitmask=AuxBitplaneAttrib.ABOGlow)

    def synthesize(self):
        return ("viewGlow",
                "pixcolor.r = pixcolor.a;\n",
                "// show the glow map by overwriting the red channel (debugging helper filter)\n")


class GammaAdjust(Filter):
    """Apply additional gamma correction to the image."""

    def __init__(self, **kwargs):
        super(GammaAdjust, self).__init__(**kwargs)

    def reset(self):
        super(GammaAdjust, self).reset()
        self.stageName = "DisplayDevice"
        self.sort = 99  # last thing in the stage; after Scanlines

    def setup(self, phase):
        # This filter needs no textures, custom inputs, aux bits, or internal stages
        pass

    def synthesize(self):
        if self.gamma == 0.5:
            code  = "// gamma = 0.5\n"
            code += "pixcolor.rgb = sqrt(pixcolor.rgb);\n"
        elif self.gamma == 2.0:
            code  = "// gamma = 2.0\n"
            code += "pixcolor.rgb *= pixcolor.rgb;\n"
        elif self.gamma != 1.0:
            code  = "// gamma = %g\n" % self.gamma
            code += "pixcolor.rgb = pow(pixcolor.rgb, %ff);\n" % self.gamma
        else: # self.gamma == 1.0:  # no-op
            code  = "// gamma = 1.0, no-op\n"

        return ("gammaAdjust",
                code,
                "// gamma correction\n")

    @property
    def gamma(self):
        """Amount of gamma correction. 1.0 = no correction."""
        return self._gamma
    @gamma.setter
    def gamma(self, value):
        if (not hasattr(self, "_gamma")  or  value != self._gamma)  and  self.pipeline is not None:
            self.pipeline._needsCompile = True
        self._gamma = value


class ColorInversion(Filter):
    """Invert the colors, like a film negative."""

    def __init__(self, **kwargs):
        super(ColorInversion, self).__init__(**kwargs)

    def reset(self):
        super(ColorInversion, self).reset()
        self.stageName = "FilmOrDetector"
        self.sort = 99

    def setup(self, phase):
        pass

    def synthesize(self):
        return ("colorInversion",
                "pixcolor = float4(1.0, 1.0, 1.0, 1.0) - pixcolor;\n",
                "// invert colors\n")


CARTOONINK_BODY="""float4 cartoondelta = %(k_cartoonseparation)s * %(texpix_txaux)s.xwyw;
float4 cartoon_c0 = tex2D(%(k_txaux)s, %(texcoord_txaux)s + cartoondelta.xy);
float4 cartoon_c1 = tex2D(%(k_txaux)s, %(texcoord_txaux)s - cartoondelta.xy);
float4 cartoon_c2 = tex2D(%(k_txaux)s, %(texcoord_txaux)s + cartoondelta.wz);
float4 cartoon_c3 = tex2D(%(k_txaux)s, %(texcoord_txaux)s - cartoondelta.wz);
float4 cartoon_mx = max(cartoon_c0, max(cartoon_c1, max(cartoon_c2, cartoon_c3)));
float4 cartoon_mn = min(cartoon_c0, min(cartoon_c1, min(cartoon_c2, cartoon_c3)));
float cartoon_thresh = saturate(dot(cartoon_mx - cartoon_mn, float4(3,3,0,0)) - 0.5);
pixcolor = lerp(pixcolor, %(k_cartooncolor)s, cartoon_thresh);
"""

class CartoonInk(Filter):
    """A cartoon outline inking filter.

    The inking is based on examining discontinuities in the normal map, as viewed from the camera.

    """
    def __init__(self, **kwargs):
        super(CartoonInk, self).__init__(**kwargs)

    def reset(self):
        super(CartoonInk, self).reset()
        # Inking must come before pretty much everything else, to simulate a completely drawn cel.
        self.stageName  = "Preprocess"
        self.sort       = 50
        self.separation = 1.0
        self.color      = (0.0, 0.0, 0.0, 1.0)

    def setup(self, phase):
        if phase == "pipeline":
            self.registerInputTexture(texName="aux")
            self.requireAuxBits(bitmask=AuxBitplaneAttrib.ABOAuxNormal)

            self.registerCustomInput(inputType="float",  inputName="k_cartoonseparation")
            self.registerCustomInput(inputType="float4", inputName="k_cartooncolor")

    def synthesize(self):
        txaux = self.getTextureInfo("aux")
        code = CARTOONINK_BODY % { "k_cartoonseparation" : self.getMangledName("k_cartoonseparation"),
                                   "k_cartooncolor"      : self.getMangledName("k_cartooncolor"),
                                   "k_txaux"             : txaux.varname,
                                   "texcoord_txaux"      : txaux.texcoord,
                                   "texpix_txaux"        : txaux.texpix }

        return ("cartoonInk",
                code,
                "// cartoon ink (outlines)\n")


    @property
    def separation(self):
        """Stencil size for examining discontinuities, in pixels.

        Pixels in the +x, -x, +y and -y directions from the current pixel
        will be examined, with the distance for each check set to 'separation'.

        Float. Values in the range 0.6 ... 1.0 are usually good. Some scenes may tolerate
        larger values; try it on your scene to see.

        """
        return self._separation
    @separation.setter
    def separation(self, value):
        self._separation = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("cartoonseparation"), value )


    @property
    def color(self):
        """Color to use for ink, as (R,G,B,A) tuple. [introduced in 1.8.0]

        The alpha component gives the _maximum_ alpha value that corresponds
        to a fully inked pixel.

        """
        return self._color
    @color.setter
    def color(self, value):
        self._color = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("cartooncolor"), Vec4(value) )

